﻿; ------------------------------------------------------------------------------------
; Stack Machine - Parser
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------

EnableExplicit

Structure _ParserLine
  RealLineCount.i ; Wo steht die Zeile in der Datei? Für debugzwecke
  ParseLine.a ; Für den 2. durchgang, wird die Zeile geparst?
  Line$ ; Zeile
EndStructure

Structure _Parser_Data
  *VTable
  Error$
  ErrorLine.i
  Map Variables.l() ; Variablen
  Map Labels.l() ; Label
  Map Constants.l() ; Konstanten
  List SourceLines._ParserLine()
EndStructure

; Parser - sInterface
Interface SM_Parser
  Parse(SM.StackMachine, Source$) ; Parst das ganze in die Stack-Machine
  GetErrorLine()
  GetError.s()
  Free()
EndInterface

; Gültige Zahl oder Constante
Procedure _Parser_IsNumeric(Map Constants.l(), Var$)
  Protected *Var.Character
  If Not Var$
    ProcedureReturn #False
  EndIf
  If FindMapElement(Constants(), LCase(Var$))
    ProcedureReturn #True
  EndIf
  Select Left(Var$, 1)
    Case "%"
      Var$ = Mid(Var$, 2)
      *Var = @Var$
      While *Var\c
        Select *Var\c  
          Case '0','1'
          Default : ProcedureReturn #False
        EndSelect
        *Var + SizeOf(Character)
      Wend
      ProcedureReturn #True
      
    Case "$"
      Var$ = Mid(Var$, 2)
      *Var = @Var$
      While *Var\c
        Select *Var\c  
          Case '0' To '9', 'a' To 'f', 'A' To 'F'
          Default : ProcedureReturn #False
        EndSelect
        *Var + SizeOf(Character)
      Wend
      ProcedureReturn #True
      
    Default
      *Var = @Var$
      While *Var\c
        Select *Var\c  
          Case '0' To '9'
          Default : ProcedureReturn #False
        EndSelect
        *Var + SizeOf(Character)
      Wend
      ProcedureReturn #True
  EndSelect
EndProcedure

Procedure _Parser_IsValidVariableName(Var$)
  Protected *Var.Character = @Var$
  If Not Var$
    ProcedureReturn #False
  EndIf
  Select *Var\c  
    Case 'a' To 'z', 'A' To 'Z', '_'
    Default : ProcedureReturn #False
  EndSelect
  *Var + SizeOf(Character)
  While *Var\c
    Select *Var\c  
      Case '0' To '9', 'a' To 'z', 'A' To 'Z', '_'
      Default : ProcedureReturn #False
    EndSelect
    *Var + SizeOf(Character)
  Wend
  ProcedureReturn #True
EndProcedure

Procedure _Parser_ShowError(*Parser._Parser_Data, ErrorLine, Error$)
  *Parser\ErrorLine = ErrorLine
  *Parser\Error$ = Error$
EndProcedure

; Konstanten ersetzen
Procedure _Parser_Val(Map Constants.l(), Value$)
  If FindMapElement(Constants(), LCase(Value$))
    ProcedureReturn Constants()
  Else
    ProcedureReturn Val(Value$)
  EndIf
EndProcedure

; Label finden
Procedure _Parser_Label(Map Label.l(), Label$, *Error.Integer)
  If FindMapElement(Label(), LCase(Label$))
    ProcedureReturn Label()
  Else
    *Error\i = #True
  EndIf
EndProcedure

; Neuer Parser
Procedure Parser_New()
  Protected *Parser._Parser_Data = AllocateMemory(SizeOf(_Parser_Data))
  InitializeStructure(*Parser, _Parser_Data)
  *Parser\VTable = ?_Parser_VTable
  ProcedureReturn *Parser
EndProcedure

; Parser freigeben
Procedure _StackMachine_Parser_Free(*Parser._Parser_Data)
  With *Parser
    ClearStructure(*Parser, _Parser_Data)
    FreeMemory(*Parser)
  EndWith
EndProcedure

; Error
Procedure _StackMachine_Parser_GetErrorLine(*Parser._Parser_Data)
  With *Parser
    ProcedureReturn \ErrorLine
  EndWith
EndProcedure

; Error
Procedure.s _StackMachine_Parser_GetError(*Parser._Parser_Data)
  With *Parser
    ProcedureReturn \Error$
  EndWith
EndProcedure

; Text in Stack-Machine einlesen
Procedure _StackMachine_ParseSource(*Parser._Parser_Data, SM.StackMachine, Source$)
  Protected CmdCount, LineCount, i, ii, Error, SubLineCount
  Protected Line$, Cmd$, Var$, Const$, SubLine$, Val.l
  
  ; Zeilentrenner umwandeln:
  Source$ = ReplaceString(Source$, Chr(10) + Chr(13), Chr(10)) + Chr(10)
  Source$ = ReplaceString(Source$, Chr(13) + Chr(10), Chr(10))
;   While FindString(Source$, Chr(10) + Chr(10))
;     Source$ = ReplaceString(Source$, Chr(10) + Chr(10), Chr(10))
;   Wend
  ; Doppelte Leerzeichen ersetzen
  Source$ = ReplaceString(Source$, Chr(9), " ")
  While FindString(Source$, "  ")
    Source$ = ReplaceString(Source$, "  ", " ")
  Wend
  
  With *Parser
    
    ; Zeilen zählen:
    LineCount = CountString(Source$, Chr(10))
    ; Einlesen in Liste...
    For i = 1 To LineCount
      ; Kommentare entfernen, alles in allem ;)
      Line$ = Trim(StringField(Trim(StringField(Source$, i, Chr(10))) + "#", 1, "#"))
      If Line$
        If FindString(Line$, ";") ; Multi-Line Command
          SubLineCount = CountString(Line$, ";")+1
          For ii = 1 To SubLineCount
            SubLine$ = Trim(StringField(Line$, ii, ";"))
            If SubLine$ ; Subzeile Ok?
              AddElement(\SourceLines())
              \SourceLines()\Line$ = SubLine$ ; Zeile speichern...
              \SourceLines()\ParseLine = #True ; Zeile parsen
              \SourceLines()\RealLineCount = i
            EndIf
          Next
        Else
          AddElement(\SourceLines())
          \SourceLines()\Line$ = Line$ ; Zeile speichern...
          \SourceLines()\ParseLine = #True ; Zeile parsen
          \SourceLines()\RealLineCount = i
        EndIf
      EndIf
    Next
    
    ; Variablen/Labels/Konstanten Erkennen und Fehlerprüfung im 1. Durchlauf
    ForEach \SourceLines()
      If Right(\SourceLines()\Line$, 1) = ":" ; Label
        ; Label-Name laden...
        Var$ = LCase(Left(\SourceLines()\Line$, Len(\SourceLines()\Line$)-1))
        If Not _Parser_IsValidVariableName(Var$) ; Fehlerhafter Name?
          ; Fehler werfen
          _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid lable name: "+\SourceLines()\Line$)
          Error = #True : Break
        EndIf
        
        ; Gibt es das Label schon?
        If AddMapElement(\Labels(), Var$)
          \Labels() = CmdCount-1 ; Label-Position speichern
          \SourceLines()\ParseLine = #False ; Line nicht mehr parsen
        Else ; Fehler, doppeltes Lable...
          ; Fehler werfen
          _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Lable already declared: "+\SourceLines()\Line$)
          Error = #True : Break
        EndIf
        
      Else ; Commando laden
        If FindString(\SourceLines()\Line$, "=") ; direkte Variablen-Deklaration
          ; Variablen-Namen und Wert laden...
          Var$ = LCase(Trim(StringField(\SourceLines()\Line$, 1, "=")))
          Val = Val(Trim(StringField(\SourceLines()\Line$, 2, "=")))
          If Not _Parser_IsValidVariableName(Var$) ; Fehlerhafter Name?
            ; Fehler werfen
            _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid variable name: "+\SourceLines()\Line$)
            Error = #True : Break
          EndIf
          
          If Not FindMapElement(\Variables(), Var$) ; Variable noch nicht gespeichert...
            AddMapElement(\Variables(), Var$) : \Variables() = Val ; Variable speichern...
          EndIf
          \SourceLines()\ParseLine = #False ; Zeile in Step 2 nicht mehr parsen...
        Else
          Cmd$ = UCase(Trim(StringField(\SourceLines()\Line$ + " ", 1, " ")))
          Var$ = LCase(Trim(StringField(\SourceLines()\Line$ + " ", 2, " ")))
          
          Select Cmd$ ; Fehlerprüfung und Variablen finden...
            Case "LDAD", "STAD", "LEA", "ADDD", "SUBD", "MULD", "DIVD", "MODD", "CMPD", "ANDD", "ORD", "XORD", "SHLD", "SHRD", "IOPTD" ; Variablen anlegen falls nötig...
              If Not _Parser_IsValidVariableName(Var$) ; Fehlerhafter Name?
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid variable name: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              
              If Not FindMapElement(\Variables(), Var$) ; Variable noch nicht gespeichert...
                AddMapElement(\Variables(), Var$) : \Variables() = #Null ; Variable speichern...
              EndIf
              
              ; Numeric - Format prüfen...
            Case "VAR", "LDAU", "ADDU", "SUBU", "MULU", "DIVU", "MODU", "CMPU", "RES", "RED", "STL", "LDL", "ADL", "SBL", "CPL", "STLI", "LDLI", "ADLI", "SBLI", "CPLI", "IOPTU", "RETN", "IF", "IFN", "ANDU", "ORU", "XORU", "SHLU", "SHRU"
              If Not _Parser_IsNumeric(\Constants(), Var$) ; Fehlerhafter Zahl?
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid numerical/constant: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              
              ; Lable-Name prüfen...
            Case "JMP", "JP", "JNP", "JN", "JNN", "JZ", "JNZ", "CALL"
              If Not _Parser_IsValidVariableName(Var$) ; Fehlerhafter Name?
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid lable name: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              
              ; Empty-Parameter prüfen
            Case "STOP", "IN", "OUT", "RET", "NEWB", "OLDB", "PUSH", "POP", "NOT"
              If Var$; Da darf keine Parameter sein...
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "No parameter allowed: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              
              ; Konstanten
            Case "CONST"
              If Not _Parser_IsValidVariableName(Var$) ; Fehlerhafter Name?
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid constant name: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              ; Konstante laden
              Const$ = LCase(Trim(StringField(\SourceLines()\Line$ + " ",3, " ")))
              
              If Not _Parser_IsNumeric(\Constants(), Const$) ; Fehlerhafter Zahl?
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Invalid numerical/constant: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              ; Konstante eintragen
              If AddMapElement(\Constants(), Var$)
                \Constants() = _Parser_Val(\Constants(), Const$)
              Else
                ; Fehler werfen
                _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Constant already declared: "+\SourceLines()\Line$)
                Error = #True : Break
              EndIf
              ; Zeile nicht mehr auswerten
              \SourceLines()\ParseLine = #False
              CmdCount - 1
              
              ; Unbekannter Befehl: 
            Default
              _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Unknown command: "+\SourceLines()\Line$)
              Error = #True : Break
              
          EndSelect
          CmdCount + 1
        EndIf
      EndIf
    Next
    
    If Not Error ; Bisher kein Fehler...
      
      ; gefundene Variablen hinten anhängen falls vorhanden
      If MapSize(\Variables())
        ; Letztes Element
        LastElement(\SourceLines())
        If UCase(\SourceLines()\Line$) <> "STOP" ; stop einfügen falls nötig
          AddElement(\SourceLines())
          \SourceLines()\Line$ = "STOP"
          \SourceLines()\ParseLine = #True
          CmdCount + 1
        EndIf
        ; Variablen anhängen
        ForEach \Variables()
          AddElement(\SourceLines())
          \SourceLines()\Line$ = "VAR "+Str(\Variables())
          \SourceLines()\ParseLine = #True
          \Variables() = CmdCount
          CmdCount + 1
        Next
      EndIf
      
      ; 2. Durchlauf: In OpCode übersetzen und fehlende Labels finden
      ForEach \SourceLines()
        If \SourceLines()\ParseLine ; Ungeparste Zeilen ausschließen
          If CountString(\SourceLines()\Line$, " ") >= 3 ; Garbage!!!!
            ; Fehler werfen
            _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Garbage at end of line: "+\SourceLines()\Line$)
            Error = #True : Break
          EndIf
          Cmd$ = UCase(StringField(\SourceLines()\Line$ + " ", 1, " "))
          Var$ = LCase(StringField(\SourceLines()\Line$ + " ", 2, " "))
          Select Cmd$
            Case "STOP" : SM\AddCommand(#OP_STOP)
            Case "VAR"  : SM\AddCommand(#OP_VAR, _Parser_Val(\Constants(), Var$))
            Case "IN"   : SM\AddCommand(#OP_IN)
            Case "OUT"  : SM\AddCommand(#OP_OUT)
            Case "IOPTU" : SM\AddCommand(#OP_IOPTU, _Parser_Val(\Constants(), Var$))
            Case "IOPTD" : SM\AddCommand(#OP_IOPTD, \Variables(Var$))
            Case "LDAU" : SM\AddCommand(#OP_LDAU, _Parser_Val(\Constants(), Var$))
            Case "LDAD" : SM\AddCommand(#OP_LDAD, \Variables(Var$))
            Case "STAD" : SM\AddCommand(#OP_STAD, \Variables(Var$))
            Case "LEA"  : SM\AddCommand(#OP_LEA, \Variables(Var$))
            Case "ADDU" : SM\AddCommand(#OP_ADDU, _Parser_Val(\Constants(), Var$))
            Case "ADDD" : SM\AddCommand(#OP_ADDD, \Variables(Var$))
            Case "SUBU" : SM\AddCommand(#OP_SUBU, _Parser_Val(\Constants(), Var$))
            Case "SUBD" : SM\AddCommand(#OP_SUBD, \Variables(Var$))
            Case "MULU" : SM\AddCommand(#OP_MULU, _Parser_Val(\Constants(), Var$))
            Case "MULD" : SM\AddCommand(#OP_MULD, \Variables(Var$))
            Case "DIVU" : SM\AddCommand(#OP_DIVU, _Parser_Val(\Constants(), Var$))
            Case "DIVD" : SM\AddCommand(#OP_DIVD, \Variables(Var$))
            Case "MODU" : SM\AddCommand(#OP_MODU, _Parser_Val(\Constants(), Var$))
            Case "MODD" : SM\AddCommand(#OP_MODD, \Variables(Var$))
            Case "ANDU" : SM\AddCommand(#OP_ANDU, _Parser_Val(\Constants(), Var$))
            Case "ANDD" : SM\AddCommand(#OP_ANDD, \Variables(Var$))
            Case "ORU"  : SM\AddCommand(#OP_ORU, _Parser_Val(\Constants(), Var$))
            Case "ORD"  : SM\AddCommand(#OP_ORD, \Variables(Var$))
            Case "XORU" : SM\AddCommand(#OP_XORU, _Parser_Val(\Constants(), Var$))
            Case "XORD" : SM\AddCommand(#OP_XORD, \Variables(Var$))
            Case "NOT"  : SM\AddCommand(#OP_NOT)
            Case "SHLU" : SM\AddCommand(#OP_SHLU, _Parser_Val(\Constants(), Var$))
            Case "SHLD" : SM\AddCommand(#OP_SHLD, \Variables(Var$))
            Case "SHRU" : SM\AddCommand(#OP_SHRU, _Parser_Val(\Constants(), Var$))
            Case "SHRD" : SM\AddCommand(#OP_SHRD, \Variables(Var$))
            Case "CMPU" : SM\AddCommand(#OP_CMPU, _Parser_Val(\Constants(), Var$))
            Case "CMPD" : SM\AddCommand(#OP_CMPD, \Variables(Var$))
            Case "JMP"  : SM\AddCommand(#OP_JMP, _Parser_Label(\Labels(), Var$, @Error))
            Case "JP"   : SM\AddCommand(#OP_JP, _Parser_Label(\Labels(), Var$, @Error))
            Case "JNP"  : SM\AddCommand(#OP_JNP, _Parser_Label(\Labels(), Var$, @Error))
            Case "JN"   : SM\AddCommand(#OP_JN, _Parser_Label(\Labels(), Var$, @Error))
            Case "JNN"  : SM\AddCommand(#OP_JNN, _Parser_Label(\Labels(), Var$, @Error))
            Case "JZ"   : SM\AddCommand(#OP_JZ, _Parser_Label(\Labels(), Var$, @Error))
            Case "JNZ"  : SM\AddCommand(#OP_JNZ, _Parser_Label(\Labels(), Var$, @Error))
            Case "IF"   : SM\AddCommand(#OP_IF, _Parser_Val(\Constants(), Var$))
            Case "IFN"  : SM\AddCommand(#OP_IFN, _Parser_Val(\Constants(), Var$))
            Case "CALL"  : SM\AddCommand(#OP_CALL, _Parser_Label(\Labels(), Var$, @Error))
            Case "RET"  : SM\AddCommand(#OP_RET)
            Case "RETN" : SM\AddCommand(#OP_RETN, _Parser_Val(\Constants(), Var$))
            Case "NEWB" : SM\AddCommand(#OP_NEWB)
            Case "OLDB" : SM\AddCommand(#OP_OLDB)
            Case "RES"  : SM\AddCommand(#OP_RES, _Parser_Val(\Constants(), Var$))
            Case "RED"  : SM\AddCommand(#OP_RED, _Parser_Val(\Constants(), Var$))
            Case "PUSH" : SM\AddCommand(#OP_PUSH)
            Case "POP"  : SM\AddCommand(#OP_POP)
            Case "STL"  : SM\AddCommand(#OP_STL, _Parser_Val(\Constants(), Var$))
            Case "LDL"  : SM\AddCommand(#OP_LDL, _Parser_Val(\Constants(), Var$))
            Case "ADL"  : SM\AddCommand(#OP_ADL, _Parser_Val(\Constants(), Var$))
            Case "SBL"  : SM\AddCommand(#OP_SBL, _Parser_Val(\Constants(), Var$))
            Case "CPL"  : SM\AddCommand(#OP_CPL, _Parser_Val(\Constants(), Var$))
            Case "STLI" : SM\AddCommand(#OP_STLI, _Parser_Val(\Constants(), Var$))
            Case "LDLI" : SM\AddCommand(#OP_LDLI, _Parser_Val(\Constants(), Var$))
            Case "ADLI" : SM\AddCommand(#OP_ADLI, _Parser_Val(\Constants(), Var$))
            Case "SBLI" : SM\AddCommand(#OP_SBLI, _Parser_Val(\Constants(), Var$))
            Case "CPLI" : SM\AddCommand(#OP_CPLI, _Parser_Val(\Constants(), Var$))
          EndSelect
          If Error ; Label-Error
            ; Fehler werfen
            _Parser_ShowError(*Parser, \SourceLines()\RealLineCount, "Label not found: "+\SourceLines()\Line$)
            Error = #True : Break
          EndIf
        EndIf 
      Next
    EndIf
    
    ; Fehler...
    If Error
      SM\Reset()
      SM\FlushMem()
      ProcedureReturn #False
    Else
      ProcedureReturn #True
    EndIf
  EndWith

EndProcedure

DataSection
  _Parser_VTable:
  Data.i @_StackMachine_ParseSource()
  Data.i @_StackMachine_Parser_GetErrorLine()
  Data.i @_StackMachine_Parser_GetError()
  Data.i @_StackMachine_Parser_Free()
EndDataSection

; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 20
; Folding = --
; EnableXP
; CPU = 1