﻿; ------------------------------------------------------------------------------------
; Stack Machine - IO Verwaltung
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------

; 32K sollte ausreichen...
#VIO_StoragePorts = 32*1024
#VIO_SpecialFunctionPorts = 32*1024

Prototype.l ExternPortHandler(StackMachine, Port.l, Akku.l)

Interface VirtualIO
  IN() ; Einlesn von IOPort
  OUT() ; An IOPort schreiben
  IOPTU(Port.l) ; IOPort wählen
  IOPTD(Port.l) ; IOPort wählen
  RegisterExternPortHandlers(ExportHandler, ImportHandler) ; Handler für externe Ports
  Free() ; Alles freigeben
EndInterface

Structure _VIO_Data
  *VTable
  *Mem._SM_Memory
  StackMachine.i
  *ImportHandler.ExternPortHandler
  *ExportHandler.ExternPortHandler
  IOPort.l ; IO-Mode
  Array Storage.l(#VIO_StoragePorts)
EndStructure


Enumeration 0
  #_IO_SF_CPUVersion ; CPU-Version
  #_IO_SF_Time ; Time
  #_IO_SF_Date ; Date
  #_IO_SF_Sleep ; Sleep
  #_IO_SF_Random ; Sleep
EndEnumeration

; Neue I/O Schnittstelle
Procedure _VIO_New(Mem, SM)
  Protected *VIO._VIO_Data = AllocateMemory(SizeOf(_VIO_Data))
  
  With *VIO
    \VTable = ?_VIO_VTable
    \Mem = Mem ; Zugriff auf den StackMachine-Memory
    \StackMachine = SM ; Für externe Handler
    InitializeStructure(*VIO, _VIO_Data)
  EndWith
  
  ProcedureReturn *VIO
EndProcedure

; Main - Input
Procedure _VIO_In(*VIO._VIO_Data)
  Protected Val.l
  With *VIO
    Select \IOPort
      Case 0 ; Manual I/0
        Val = Val(InputRequester("Input", "Input to Akku from Port 0 ", "0"))
        
      Case 1 To #VIO_StoragePorts ; Storage-Ports
        Val = \Storage(\IOPort)
        
      Case #VIO_StoragePorts+1 To #VIO_StoragePorts+#VIO_SpecialFunctionPorts ; Special-Function Ports
        Select (\IOPort-(#VIO_StoragePorts+1))
          Case #_IO_SF_CPUVersion
            Val = #SM_CPU_Version
          Case #_IO_SF_Time
            Val = ElapsedMilliseconds()
          Case #_IO_SF_Date
            Val = Date()
          Case #_IO_SF_Random
            Val = Random(2147483647)
            
        EndSelect
        
      Default
        If \ImportHandler
          Val = \ImportHandler(\StackMachine, \IOPort, 0)
        EndIf
        
    EndSelect
    
    ; In Akku schreiben
    \Mem\Akku\Set(Val)
  EndWith
EndProcedure

; Main - Output
Procedure _VIO_Out(*VIO._VIO_Data)
  Protected Val.l
  With *VIO
    Val = \Mem\Akku\Get()
    Select \IOPort
      Case 0 ; Manual I/0
        MessageRequester("", "AKKU to Port 0"+": "+ Str(Val))
        
      Case 1 To #VIO_StoragePorts ; Storage-Ports
        \Storage(\IOPort) = Val
        
      Case #VIO_StoragePorts+1 To #VIO_StoragePorts+#VIO_SpecialFunctionPorts ; Special-Function Ports
        Select (\IOPort-(#VIO_StoragePorts+1))
          Case #_IO_SF_Sleep
            Delay(Val) ; Warten...

        EndSelect
        
      Default
        If \ExportHandler
          \ExportHandler(\StackMachine, \IOPort, Val)
        EndIf
    EndSelect
  EndWith
EndProcedure

; Main - Output festlegen
Procedure _VIO_IOPTU(*VIO._VIO_Data, IOPort.l)
  Protected Val.l
  With *VIO
    \IOPort = IOPort
  EndWith
EndProcedure

Procedure _VIO_IOPTD(*VIO._VIO_Data, IOPort.l)
  Protected Val.l
  With *VIO
    \IOPort = \Mem\Ram\Get(IOPort)
  EndWith
EndProcedure

Procedure _VIO_RegisterExternPortHandlers(*VIO._VIO_Data, ExportHandler, ImportHandler)
  Protected Val.l
  With *VIO
    \ImportHandler = ImportHandler
    \ExportHandler = ExportHandler
  EndWith
EndProcedure

; I/O freigeben
Procedure _VIO_Free(*VIO._VIO_Data)
  With *VIO
    ClearStructure(*VIO, _VIO_Data)
    FreeMemory(*VIO)
  EndWith
EndProcedure

DataSection
  _VIO_VTable:
  Data.i @_VIO_IN()
  Data.i @_VIO_OUT()
  Data.i @_VIO_IOPTU()
  Data.i @_VIO_IOPTD()
  Data.i @_VIO_RegisterExternPortHandlers()
  Data.i @_VIO_Free()
EndDataSection
; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 75
; FirstLine = 54
; Folding = --
; EnableXP
; CPU = 1