﻿; ------------------------------------------------------------------------------------
; Stack Machine - Virtual Processor
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------


Structure _VProc_Data
  *VTable
  *Mem._SM_Memory
EndStructure

Interface VirtualProcessor
  Free()
  LDAU(n)
  LDAD(n)
  STAD(n)
  LEA(n)
  ADDU(n)
  ADDD(n)
  SUBU(n)
  SUBD(n)
  MULU(n)
  MULD(n)
  DIVU(n)
  DIVD(n)
  MODU(n)
  MODd(n)
  AndU(n)
  AndD(n)
  OrU(n)
  OrD(n)
  XOrU(n)
  XOrD(n)
  Not(n)
  SHLU(n)
  SHLD(n)
  SHRU(n)
  SHRD(n)
  CMPU(n)
  CMPD(n)
  JUM(n)
  JP(n)
  JNP(n)
  JN(n)
  JNN(n)
  JZ(n)
  JNZ(n)
  If(n)
  IFN(n)
  CAL(n)
  RET(n)
  RETN(n)
  NEWB(n)
  OLDB(n)
  RES(n)
  Red(n)
  PUSH(n)
  POP(n)
  STL(n)
  LDL(n)
  ADL(n)
  SBL(n)
  CPL(n)
  STLI(n)
  LDLI(n)
  ADLI(n)
  SBLI(n)
  CPLI(n)
EndInterface

; Neuen Virtuellen Prozessor erstellen
Procedure _VProc_New(*SMM._SM_Memory)
  Protected *VP._VProc_Data = AllocateMemory(SizeOf(_VProc_Data))
  With *VP
    \VTable = ?_VProc_VTable
    \Mem = *SMM
  EndWith
  
  ProcedureReturn *VP
EndProcedure

Procedure _VirtualProcessor_Free(*VP._VProc_Data)
  FreeMemory(*VP)
EndProcedure

; Befehle:
Macro _VP_JmpCmdX(_CmdName, _Cmp, _CmdLine)
  Procedure _VirtualProcessor_Cmd_#_CmdName(*VP._VProc_Data, n)
    If _Cmp
      _CmdLine
    EndIf
  EndProcedure
EndMacro

Macro _VP_CmdX(_CmdName, _CmdLine)
  Procedure _VirtualProcessor_Cmd_#_CmdName(*VP._VProc_Data, n)
    _CmdLine
  EndProcedure
EndMacro

; Store and Read Data
_VP_CmdX(LDAU, *VP\Mem\Akku\Set(n))
_VP_CmdX(LDAD, *VP\Mem\Akku\Set(*VP\Mem\Ram\Get(n)))
_VP_CmdX(STAD, *VP\Mem\Ram\Set(n, *VP\Mem\Akku\Get()))
_VP_CmdX(LEA, *VP\Mem\Akku\Set(n))

; Arithmetic:
_VP_CmdX(ADDU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() + n))
_VP_CmdX(ADDD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() + *VP\Mem\Ram\Get(n)))
_VP_CmdX(SUBU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() - n))
_VP_CmdX(SUBD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(n)))
_VP_CmdX(MULU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() * n))
_VP_CmdX(MULD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() * *VP\Mem\Ram\Get(n)))
_VP_CmdX(DIVU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() / n))
_VP_CmdX(DIVD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() / *VP\Mem\Ram\Get(n)))
_VP_CmdX(MODU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() % n))
_VP_CmdX(MODD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() % *VP\Mem\Ram\Get(n)))
_VP_CmdX(ANDU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() & n))
_VP_CmdX(ANDD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() & *VP\Mem\Ram\Get(n)))
_VP_CmdX(ORU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() | n))
_VP_CmdX(ORD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() | *VP\Mem\Ram\Get(n)))
_VP_CmdX(XORU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() ! n))
_VP_CmdX(XORD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() ! *VP\Mem\Ram\Get(n)))
_VP_CmdX(Not, *VP\Mem\Akku\Set(~*VP\Mem\Akku\Get()))
_VP_CmdX(SHLU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() << n))
_VP_CmdX(SHLD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() << *VP\Mem\Ram\Get(n)))
_VP_CmdX(SHRU, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() >> n))
_VP_CmdX(SHRD, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() >> *VP\Mem\Ram\Get(n)))

; Comparison:
_VP_CmdX(CMPU, *VP\Mem\Result\Set(*VP\Mem\Akku\Get() - n))
_VP_CmdX(CMPD, *VP\Mem\Result\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(n)))

;Command Pointer:
_VP_JmpCmdX(JUM, #True, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JP, *VP\Mem\Result\Get() > #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JNP, *VP\Mem\Result\Get() <= #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JN, *VP\Mem\Result\Get() < #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JNN, *VP\Mem\Result\Get() >= #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JZ, *VP\Mem\Result\Get() = #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(JNZ, *VP\Mem\Result\Get() <> #Null, *VP\Mem\CP\Set(n))
_VP_JmpCmdX(If, *VP\Mem\Result\Get() = #Null, *VP\Mem\CP\Set(*VP\Mem\CP\Get() + n))
_VP_JmpCmdX(IFN, *VP\Mem\Result\Get() <> #Null, *VP\Mem\CP\Set(*VP\Mem\CP\Get() + n))

; Procedure Commands:
Procedure _VirtualProcessor_Cmd_CAL(*VP._VProc_Data, n)
  With *VP\Mem
    \Ram\Set(\SP\Get(), \CP\Get())
    \CP\Set(n)
    \SP\Dec()
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_RET(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Inc()
    \CP\Set(\Ram\Get(\SP\Get()))
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_RETN(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Inc()
    \CP\Set(\Ram\Get(\SP\Get()))
    If n >= 0
      \SP\Set(\SP\Get() + n)
    Else ; Interrupt Return
      \InterruptLevel\Dec()
    EndIf
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_NEWB(*VP._VProc_Data, n)
  With *VP\Mem
    \Ram\Set(\SP\Get(), \BP\Get())
    \BP\Set(\SP\Get())
    \SP\Dec()
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_OLDB(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Inc()
    \BP\Set(\Ram\Get(\SP\Get()))
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_RES(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Set(\SP\Get() - n)
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_RED(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Set(\SP\Get() + n)
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_PUSH(*VP._VProc_Data, n)
  With *VP\Mem
    \Ram\Set(\SP\Get(), \Akku\Get())
    \SP\Dec()
  EndWith
EndProcedure
Procedure _VirtualProcessor_Cmd_POP(*VP._VProc_Data, n)
  With *VP\Mem
    \SP\Inc()
    \Akku\Set(\Ram\Get(\SP\Get()))
  EndWith
EndProcedure

; Stack Commands by Value:
_VP_CmdX(STL, *VP\Mem\Ram\Set(*VP\Mem\BP\Get() + n, *VP\Mem\Akku\Get()))
_VP_CmdX(LDL, *VP\Mem\Akku\Set(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n)))
_VP_CmdX(ADL, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() + *VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n)))
_VP_CmdX(SBL, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n)))
_VP_CmdX(CPL, *VP\Mem\Result\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n)))

; Stack Commands by Reference:
_VP_CmdX(STLI, *VP\Mem\Ram\Set(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n), *VP\Mem\Akku\Get()))
_VP_CmdX(LDLI, *VP\Mem\Akku\Set(*VP\Mem\Ram\Get(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n))))
_VP_CmdX(ADLI, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() + *VP\Mem\Ram\Get(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n))))
_VP_CmdX(SBLI, *VP\Mem\Akku\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n))))
_VP_CmdX(CPLI, *VP\Mem\Result\Set(*VP\Mem\Akku\Get() - *VP\Mem\Ram\Get(*VP\Mem\Ram\Get(*VP\Mem\BP\Get() + n))))


DataSection
  _VProc_VTable:
  Data.i @_VirtualProcessor_Free()
  Data.i @_VirtualProcessor_Cmd_LDAU()
  Data.i @_VirtualProcessor_Cmd_LDAD()
  Data.i @_VirtualProcessor_Cmd_STAD()
  Data.i @_VirtualProcessor_Cmd_LEA()
  Data.i @_VirtualProcessor_Cmd_ADDU()
  Data.i @_VirtualProcessor_Cmd_ADDD()
  Data.i @_VirtualProcessor_Cmd_SUBU()
  Data.i @_VirtualProcessor_Cmd_SUBD()
  Data.i @_VirtualProcessor_Cmd_MULU()
  Data.i @_VirtualProcessor_Cmd_MULD()
  Data.i @_VirtualProcessor_Cmd_DIVU()
  Data.i @_VirtualProcessor_Cmd_DIVD()
  Data.i @_VirtualProcessor_Cmd_MODU()
  Data.i @_VirtualProcessor_Cmd_MODD()
  Data.i @_VirtualProcessor_Cmd_ANDU()
  Data.i @_VirtualProcessor_Cmd_ANDD()
  Data.i @_VirtualProcessor_Cmd_ORU()
  Data.i @_VirtualProcessor_Cmd_ORD()
  Data.i @_VirtualProcessor_Cmd_XORU()
  Data.i @_VirtualProcessor_Cmd_XORD()
  Data.i @_VirtualProcessor_Cmd_NOT()
  Data.i @_VirtualProcessor_Cmd_SHLU()
  Data.i @_VirtualProcessor_Cmd_SHLD()
  Data.i @_VirtualProcessor_Cmd_SHRU()
  Data.i @_VirtualProcessor_Cmd_SHRD()
  Data.i @_VirtualProcessor_Cmd_CMPU()
  Data.i @_VirtualProcessor_Cmd_CMPD()
  Data.i @_VirtualProcessor_Cmd_JUM()
  Data.i @_VirtualProcessor_Cmd_JP()
  Data.i @_VirtualProcessor_Cmd_JNP()
  Data.i @_VirtualProcessor_Cmd_JN()
  Data.i @_VirtualProcessor_Cmd_JNN()
  Data.i @_VirtualProcessor_Cmd_JZ()
  Data.i @_VirtualProcessor_Cmd_JNZ()
  Data.i @_VirtualProcessor_Cmd_IF()
  Data.i @_VirtualProcessor_Cmd_IFN()
  Data.i @_VirtualProcessor_Cmd_CAL()
  Data.i @_VirtualProcessor_Cmd_RET()
  Data.i @_VirtualProcessor_Cmd_RETN()
  Data.i @_VirtualProcessor_Cmd_NEWB()
  Data.i @_VirtualProcessor_Cmd_OLDB()
  Data.i @_VirtualProcessor_Cmd_RES()
  Data.i @_VirtualProcessor_Cmd_Red()
  Data.i @_VirtualProcessor_Cmd_PUSH()
  Data.i @_VirtualProcessor_Cmd_POP()
  Data.i @_VirtualProcessor_Cmd_STL()
  Data.i @_VirtualProcessor_Cmd_LDL()
  Data.i @_VirtualProcessor_Cmd_ADL()
  Data.i @_VirtualProcessor_Cmd_SBL()
  Data.i @_VirtualProcessor_Cmd_CPL()
  Data.i @_VirtualProcessor_Cmd_STLI()
  Data.i @_VirtualProcessor_Cmd_LDLI()
  Data.i @_VirtualProcessor_Cmd_ADLI()
  Data.i @_VirtualProcessor_Cmd_SBLI()
  Data.i @_VirtualProcessor_Cmd_CPLI()
EndDataSection
; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 165
; FirstLine = 141
; Folding = ---
; EnableXP
; CPU = 1