﻿; ------------------------------------------------------------------------------------
; Stack Machine - VRam Modul
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------

Interface VirtualRam
  GetOpCode(Index) ; OpCode aus Ram laden
  Get.l(Index) ; Long aus Ram laden
  SetOpCode(Index, *OP.OpCode) ; OpCode schreiben
  Set(Index, Value.l) ; Long aus Ram laden
  GetSize.l() ; Gesamt-Größe des Ram ermitteln
  Flush() ; Ram Leeren
  Free() ; Ram freigeben
EndInterface

Structure _VirtualRam_Data
  *VTable
  *Mem
  MemSize.i
EndStructure

Procedure _VRam_New(Size)
  Protected *VR._VirtualRam_Data = AllocateMemory(SizeOf(_VirtualRam_Data))
  With *VR
    \Mem = AllocateMemory(Size)
    \MemSize = Size
    \VTable = ?_VRam_VTable
  EndWith
  
  ProcedureReturn *VR
EndProcedure

Procedure _VRam_GetOpCode(*VR._VirtualRam_Data, Index)
  With *VR
    ProcedureReturn *VR\Mem + Index*SizeOf(OpCode)
  EndWith
EndProcedure

Procedure.l _VRam_Get(*VR._VirtualRam_Data, Index)
  With *VR
    ProcedureReturn PeekL(*VR\Mem + Index*SizeOf(OpCode) + SizeOf(Ascii))
  EndWith
EndProcedure

Procedure _VRam_SetOpCode(*VR._VirtualRam_Data, Index, *OP)
  With *VR
    CopyMemory(*OP, *VR\Mem + Index*SizeOf(OpCode), SizeOf(OpCode))
  EndWith
EndProcedure

Procedure _VRam_Set(*VR._VirtualRam_Data, Index, Value.l)
  With *VR
    ProcedureReturn PokeL(*VR\Mem + Index*SizeOf(OpCode) + SizeOf(Ascii), Value)
  EndWith
EndProcedure

Procedure.l _VRam_GetSize(*VR._VirtualRam_Data)
  With *VR
    ProcedureReturn *VR\MemSize
  EndWith
EndProcedure

Procedure _VRam_Flush(*VR._VirtualRam_Data)
  With *VR
    FillMemory(\Mem, \MemSize)
  EndWith
EndProcedure

Procedure _VRam_Free(*VR._VirtualRam_Data)
  With *VR
    FreeMemory(\Mem)
    FreeMemory(*VR)
  EndWith
EndProcedure

DataSection
  _VRam_VTable:
  Data.i @_VRam_GetOpCode()
  Data.i @_VRam_Get()
  Data.i @_VRam_SetOpCode()
  Data.i @_VRam_Set()
  Data.i @_VRam_GetSize()
  Data.i @_VRam_Flush()
  Data.i @_VRam_Free()
EndDataSection


; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 14
; Folding = --
; EnableXP
; CPU = 1